"""zadeh - Python package to build fuzzy inference systems"""

__version__ = '0.1.0'
__author__ = 'Dih5 <dihedralfive@gmail.com>'
__all__ = []

from .domains import *
from .sets import *
from .variables import *
from .rules import *
from .fis import *
